/*
DO NOT MODIFY THIS FILE.  It is automatically generated and changes will be over-written
on the next generation.

It contains enumerations used by the {{stinger.name}} interface.
*/

{%for prop_name, prop in stinger.properties.items() %}
// {{prop}}
{%endfor%}