/*
DO NOT MODIFY THIS FILE.  It is automatically generated and changes will be over-written
on the next generation.

It contains enumerations used by the {{stinger.name}} interface.
*/

#pragma once
#include <string>
#include <boost/optional.hpp>
#include "{{stinger.cpp.enum_header_file}}"

{%for istruct_name, istruct in stinger.structs.items() %}
struct {{istruct_name | UpperCamelCase }} {
    {%-for value in istruct.members %}
    {{value.cpp_temp_type}} {{value.name}};
    {%-endfor%}
};
{%endfor%}
