[package]
name = "{{stinger.rust.common_package_name}}"
version = "{{stinger.version}}"
edition = "2024"

[dependencies]
futures = "0.3.31"
mqttier = { git = "https://github.com/stinger-ipc/mqttier.git" }
serde = { version = "1.0.219", features = ["derive"] }
serde_json = "1.0.142"
tokio = { version = "1", features = ["full"] }
num-derive = "0.4.2"
num-traits = "0.2.19"

[dependencies.uuid]
version = "1.17.0"
features = [
    "v4",                # Lets you generate random UUIDs
    "fast-rng",          # Use a faster (but still sufficiently random) RNG
    "macro-diagnostics", # Enable better diagnostics for compile-time UUIDs
]

[[example]]
name = "{{stinger.name | snake_case}}_connection_demo"
path = "examples/pub_and_recv.rs"